#define _GNU_SOURCE
#include "window.hh"
#include "icon.hh"


/*

  Resources allocated for each icon:

  Debugger - debug
  Window - icon window, label (no context), subwindow
  GC - iconGC
  XPointer - icon name,
*/

BlackboxIcon::BlackboxIcon(BlackboxSession *ctrl, XID c, XID i, XID m,
			   Bool use_iconwin) {
  session = ctrl;
  display = session->control();

  debug = new Debugger('$');
#ifdef DEBUG
  debug->enable();
#endif
  debug->enter("BlackboxIcon::BlackboxIcon %lx\n", m);

  icon.client = c;
  icon.label = None;
  moving = False;

  use_icon_window = use_iconwin;
  shaped = (icon.mask != None && session->shapeExtensions());

  unsigned long attrib_mask = CWBackPixmap|CWBackPixel|CWBorderPixel|
    CWCursor|CWEventMask; 
  XSetWindowAttributes attrib;
  attrib.background_pixmap = None;
  attrib.background_pixel = session->iconPixel();
  attrib.border_pixel = 0l;
  attrib.cursor = session->sessionCursor();
  attrib.event_mask = StructureNotifyMask|SubstructureNotifyMask|
    ExposureMask|SubstructureRedirectMask|ButtonPressMask|
    ButtonReleaseMask|ButtonMotionMask|ExposureMask;
  
  icon.window =
    XCreateWindow(display, session->Root(), 0, 0, 1, 1, 0, session->Depth(),
		  InputOutput, session->visual(), attrib_mask, &attrib);
  XSaveContext(display, icon.window, session->iconContext(),
	       (XPointer) this);
  debug->msg("creating icon window %lx %lx\n", icon.window, icon.client);
  
  /* create our label that contains the icon text */
  attrib.border_pixel = session->framePixel();
  icon.label =
    XCreateWindow(display, session->Root(), 0, 0, 1, 1, 1, session->Depth(),
		  InputOutput, session->visual(), attrib_mask, &attrib);

  XGCValues gcv;
  gcv.foreground = session->getColor("white");
  gcv.font = session->titleFont()->fid;
  iconGC = XCreateGC(display, icon.window, GCForeground|GCFont, &gcv);

  if (! XGetIconName(display, icon.client, &icon.name))
    icon.name = NULL;

  int namewidth = (icon.name != NULL) ? 
    XTextWidth(session->titleFont(), icon.name, strlen(icon.name)) :
    XTextWidth(session->titleFont(), "Untitled", strlen("Untitled"));

  icon.width_l = namewidth + 4;
  icon.height_l = session->titleFont()->ascent +
    session->titleFont()->descent + 4;
  XResizeWindow(display, icon.label, icon.width_l, icon.height_l);

  icon.subwindow = None;

  //
  // for now... we use one pixmap for icons... support for program specified
  // icons is planned.
  //

  if ((! use_icon_window) || (use_icon_window && (i == None))) {
    icon.pixmap = session->defaultIcon();
    icon.mask = session->defaultIconMask();
    icon.width = session->defaultIconAttributes()->width;
    icon.height = session->defaultIconAttributes()->height;
    
    attrib.event_mask |= EnterWindowMask|LeaveWindowMask;
    icon.subwindow =
      XCreateWindow(display, icon.window, 0, 0, icon.width, icon.height, 0,
		    session->Depth(), InputOutput, session->visual(),
		    attrib_mask, &attrib);
    XSetWindowBackgroundPixmap(display, icon.subwindow, icon.pixmap);
    XSaveContext(display, icon.subwindow, session->iconContext(),
		 (XPointer) this);

    debug->msg("positioning icon %lx %lx %lx\n", icon.window, icon.subwindow,
	       icon.pixmap);

    XMoveResizeWindow(display, icon.window,
		      session->XResolution() - icon.width - 2,
		      (icon.height + 2) * session->iconCount(),
		      icon.width, icon.height);
    
#ifdef SHAPE
    if (icon.mask != None) {
      debug->msg("shaping icon\n");
      XShapeCombineMask(display, icon.window, ShapeSet, 0, 0, icon.mask,
			ShapeBounding);
    }
#endif
    
    debug->msg("mapping icon\n");
    XMapSubwindows(display, icon.window);
    XMapRaised(display, icon.window);
    session->addIcon(this);
  } else
    debug->msg("using icon window\n");
}



/*

  Resources deallocated for each icon:

  Debugger - debug
  Window - icon window, label (no context), subwindow
  GC - iconGC
  XPointer - icon name,
*/

BlackboxIcon::~BlackboxIcon(void) {
  /*
    remove icon from root window
  */
  XUnmapWindow(display, icon.window);
  XUnmapWindow(display, icon.label);

  /*
    dissociate this icon from the controlling session
  */
  debug->msg("deleting icon\n");
  session->removeIcon(this);
  session->arrangeIcons();

  XFreeGC(display, iconGC);
  XDeleteContext(display, icon.window, session->iconContext());
  XDeleteContext(display, icon.subwindow, session->iconContext());
  if (icon.name != NULL) XFree(icon.name);
  if (icon.subwindow != None) XDestroyWindow(display, icon.subwindow);
  XDestroyWindow(display, icon.window);
  XDestroyWindow(display, icon.label);

  delete debug;
}


void BlackboxIcon::buttonPressEvent(XButtonEvent *) {
  debug->msg("icon pressed\n");
  if (session->button1Pressed())
    XRaiseWindow(display, icon.window);
}


void BlackboxIcon::buttonReleaseEvent(XButtonEvent *be) {
  debug->msg("icon released\n");
  if (be->button == 1) {
    if (! moving) {
      XGrabServer(display);
      debug->msg("deiconifying app\n");
      BlackboxWindow *win = session->getWindow(display, icon.client);
      XUnmapWindow(display, icon.label);
      XUnmapWindow(display, icon.window);
      XUngrabServer(display);
      win->deiconifyWindow();
    } else {
      debug->msg("ending icon move\n");
      XUngrabPointer(display, CurrentTime);
      moving = False;
    }
  }
}


void BlackboxIcon::motionNotifyEvent(XMotionEvent *me) {
  if (session->button1Pressed())
    if (! moving)
      if (XGrabPointer(display, icon.window, False, PointerMotionMask|
		       ButtonReleaseMask, GrabModeAsync, GrabModeAsync,
		       None, session->moveCursor(), CurrentTime)
	  == GrabSuccess) {
	moving = 1;
	icon.x_move = me->x;
	icon.y_move = me->y;
	debug->msg("starting icon move %d, %d\n", me->x, me->y);
      } else {
	moving = 0;
	debug->msg("couldn't grab pointer for icon move\n");
      }
    else {
      int dx = me->x_root - icon.x_move,
	dy = me->y_root - icon.y_move;
      XMoveWindow(display, icon.window, dx, dy);
    }
}


void BlackboxIcon::enterNotifyEvent(XCrossingEvent *) {
  XWindowAttributes att;
  XGetWindowAttributes(display, icon.window, &att);

  int lx, ly;
  lx = att.x - icon.width_l;
  ly = (att.y + (att.height / 2)) - (icon.height_l / 2);
  XMoveWindow(display, icon.label, lx, ly);
  XMapRaised(display, icon.label);
  XDrawString(display, icon.label, iconGC, 2, 2 + session->titleFont()->ascent,
	      ((icon.name != NULL) ? icon.name : "Untitled"),
	      ((icon.name != NULL) ? strlen(icon.name) : strlen("Untitled")));
}


void BlackboxIcon::leaveNotifyEvent(XCrossingEvent *) {
  XUnmapWindow(display, icon.label);
}
